﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Algorithms.PushGP.Expressions;

namespace HeuristicLab.Algorithms.PushGP
{
    public static class ExpressionTable
    {
        private static Dictionary<string, Expression> expressionTable;
        private static Dictionary<string, Func<Expression>> statefullExpressionFactory;

        public static int Count;

        static ExpressionTable()
        {
            expressionTable = GetExperssionTable(e => e.ToString());
            statefullExpressionFactory = GetStatefullExpressionFactory(e => e.ToString());

            Count = expressionTable.Count + statefullExpressionFactory.Count;
        }

        private static Dictionary<string, Expression> GetExperssionTable(Func<Expression, string> keyFunc)
        {
            return GetExperssionTable(e => new KeyValuePair<string, Expression>(keyFunc(e), e));
        }

        private static Dictionary<string, Expression> GetExperssionTable(Func<Expression, KeyValuePair<string, Expression>> pairFunc)
        {
            var dictionary = new Dictionary<string, Expression>();
            var expressionTypes = GetExpressionTypes(filterStatefull: false);

            foreach (var type in expressionTypes)
            {
                var expression = (Expression)Activator.CreateInstance(type);
                var pair = pairFunc(expression);

                dictionary.Add(pair.Key, pair.Value);
            }

            return dictionary;
        }

        private static Dictionary<string, Func<Expression>> GetStatefullExpressionFactory(Func<Expression, string> keyFunc)
        {
            var dictionary = new Dictionary<string, Func<Expression>>();
            var expressionTypes = GetExpressionTypes(filterStatefull: true);

            foreach (var type in expressionTypes)
            {
                // Make a NewExpression that calls the ctor
                var newExp = System.Linq.Expressions.Expression.New(type);

                // Create a lambda with the New expression as body
                var creator = System.Linq.Expressions.Expression.Lambda<Func<Expression>>(newExp).Compile();

                var expression = creator();
                var key = keyFunc(expression);

                dictionary.Add(key, creator);
            }

            return dictionary;
        }

        private static IEnumerable<Type> GetExpressionTypes(bool filterStatefull)
        {
            return from domainAssembly in AppDomain.CurrentDomain.GetAssemblies()
                   from assemblyType in domainAssembly.GetTypes()
                   let type = typeof(Expression)
                   where !assemblyType.IsAbstract &&
                         assemblyType.IsSubclassOf(type) &&
                         Attribute.IsDefined(assemblyType, typeof(StatefullExpressionAttribute)) == filterStatefull &&
                         assemblyType.GetConstructor(Type.EmptyTypes) != null
                   select assemblyType;
        }

        public static int StatelessCount { get { return expressionTable.Count; } }

        public static int StatefullCount { get { return statefullExpressionFactory.Count; } }

        public static Expression GetExpression(int index)
        {
            return index < StatelessCount
                ? expressionTable.ElementAt(index).Value
                : statefullExpressionFactory.ElementAt(index - StatelessCount).Value();
        }

        public static Expression GetExpression(string symbol)
        {
            Expression expression;
            if (TryGetStatelessExpression(symbol, out expression) ||
                TryGetStatefullExpression(symbol, out expression))
            {
                return expression;
            }

            return null;
        }

        public static Expression GetStatelessExpression(string symbol)
        {
            Expression expression;
            if (expressionTable.TryGetValue(symbol, out expression)) return expression;
            else throw new NotSupportedException("Expression not supported: " + symbol);
        }

        public static bool TryGetStatelessExpression(string name, out Expression expression)
        {
            return expressionTable.TryGetValue(name, out expression);
        }

        public static Expression GetStatefullExpression(string symbol)
        {
            Func<Expression> creator;
            if (statefullExpressionFactory.TryGetValue(symbol, out creator)) return creator();
            else throw new NotSupportedException("Expression not supported: " + symbol);
        }

        public static bool TryGetStatefullExpression(string name, out Expression expression)
        {
            Func<Expression> creator;
            if (statefullExpressionFactory.TryGetValue(name, out creator))
            {
                expression = creator();
                return true;
            }
            else
            {
                expression = default(Expression);
                return false;
            }

        }
    }
}
