﻿using System;
using HeuristicLab.Algorithms.PushGP.Interpreter;

namespace HeuristicLab.Algorithms.PushGP.Expressions
{
    /// <summary>
    /// Pushes the sum of the top two items.
    /// </summary>
    public class FloatAddExpression : PushResultExpression<double>
    {
        public FloatAddExpression() : base("FLOAT.+")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack, 2, values => values[0] + values[1]);
        }
    }

    /// <summary>
    /// Pushes the difference of the top two items; that is, the second item minus the top item.
    /// </summary>
    public class FloatSubtractExpression : PushResultExpression<double>
    {
        public FloatSubtractExpression() : base("FLOAT.-")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack, 2, values => values[0] - values[1]);
        }
    }

    /// <summary>
    /// Pushes the product of the top two items.
    /// </summary>
    public class FloatMultiplyExpression : PushResultExpression<double>
    {
        public FloatMultiplyExpression() : base("FLOAT.*")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack, 2, values => values[0] * values[1]);
        }
    }

    /// <summary>
    /// Pushes the quotient of the top two items; that is, the second item divided by the top item.
    /// If the top item is zero this acts as a NOOP.
    /// </summary>
    public class FloatDivideExpression : PushResultExpression<double>
    {
        public FloatDivideExpression() : base("FLOAT./")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack, 2, values => values[0] / values[1], 0);
        }
    }

    /// <summary>
    /// Pushes the second stack item modulo the top stack item. If the top item is zero this acts as a NOOP.
    /// The modulus is computed as the remainder of the quotient, where the quotient has first been truncated toward negative infinity.
    /// (This is taken from the definition for the generic MOD function in Common Lisp, which is described
    /// for example at http://www.lispworks.com/reference/HyperSpec/Body/f_mod_r.htm.)
    /// </summary>
    public class FloatModuloExpression : PushResultExpression<double>
    {
        public FloatModuloExpression() : base("FLOAT.%")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack, 2, values => values[0] % values[1], 0);
        }
    }

    /// <summary>
    /// Pushes the minimum of the top two items.
    /// </summary>
    public class FloatMinExpression : PushResultExpression<double>
    {
        public FloatMinExpression() : base("FLOAT.MIN")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack, 2, values => Math.Min(values[0], values[1]));
        }
    }

    /// <summary>
    /// Pushes the maximum of the top two items.
    /// </summary>
    public class FloatMaxExpression : PushResultExpression<double>
    {
        public FloatMaxExpression() : base("FLOAT.MAX")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack, 2, values => Math.Max(values[0], values[1]));
        }
    }

    /// <summary>
    /// Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
    /// </summary>
    public class FloatSmallerThanExpression : PushResultExpression<double>
    {
        public FloatSmallerThanExpression() : base("FLOAT.<")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(
                interpreter.FloatStack,
                interpreter.BooleanStack,
                2,
                values => values[0] < values[1]);
        }
    }

    /// <summary>
    /// Pushes TRUE onto the BOOLEAN stack if the second item is greater than the top item, or FALSE otherwise.
    /// </summary>
    public class FloatGreaterThanExpression : PushResultExpression<double>
    {
        public FloatGreaterThanExpression() : base("FLOAT.>")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(
                interpreter.FloatStack,
                interpreter.BooleanStack,
                2,
                values => values[0] > values[1]);
        }
    }

    /// <summary>
    /// Pushes the sine of the top item.
    /// </summary>
    public class FloatSineExpression : PushResultExpression<double>
    {
        public FloatSineExpression() : base("FLOAT.SIN")
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(
                interpreter.FloatStack,
                1,
                values => Math.Sin(values[0]));
        }
    }

    /// <summary>
    /// Pushes the cosine of the top item.
    /// </summary>
    public class FloatCosineExpression : PushResultExpression<double>
    {
        public FloatCosineExpression() : base("FLOAT.COS") { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(
                interpreter.FloatStack,
                1,
                values => Math.Cos(values[0]));
        }
    }

    /// <summary>
    /// Pushes 1 if the top BOOLEAN is TRUE, or 0 if the top BOOLEAN is FALSE.
    /// </summary>
    public class FloatFromBooleanExpression : Expression
    {
        public FloatFromBooleanExpression() : base("FLOAT.FROMBOOLEAN") { }

        public override void Eval(IInterpreter interpreter)
        {
            if (interpreter.BooleanStack.Count == 0)
                return;

            var condition = interpreter.BooleanStack.Pop();
            var value = condition ? 1 : 0;

            interpreter.FloatStack.Push(value);
        }
    }

    /// <summary>
    /// Pushes a floating point version of the top INTEGER.
    /// </summary>
    public class FloatFromIntegerExpression : Expression
    {
        public FloatFromIntegerExpression() : base("FLOAT.FROMINTEGER")
        {
        }

        public override void Eval(IInterpreter interpreter)
        {
            if (interpreter.IntegerStack.Count == 0)
                return;

            var value = (double)interpreter.IntegerStack.Pop();

            interpreter.FloatStack.Push(value);
        }
    }
}
