﻿using HeuristicLab.Algorithms.PushGP.Interpreter;
using HeuristicLab.Algorithms.PushGP.Stack;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.PushGP.Expressions
{
    public abstract class PushExpression<T> : Expression
    {
        protected readonly T value;

        public PushExpression(T value) : base(value.ToString(), value.GetHashCode())
        {
            this.value = value;
        }

        // copy constructor
        public PushExpression(PushExpression<T> expression, Cloner cloner) : base(expression, cloner)
        {
            this.value = expression.value;
        }

        protected void Eval(IStack<T> stack)
        {
            stack.Push(this.value);
        }

        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is PushExpression<T>))
            {
                return false;
            }

            var other = obj as PushExpression<T>;

            return this.value.Equals(value);
        }
    }

    public class IntegerPushExpression : PushExpression<long>
    {
        public IntegerPushExpression(long value) : base(value)
        { }

        // copy constructor
        public IntegerPushExpression(IntegerPushExpression expression, Cloner cloner) : base(expression, cloner)
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack);
        }

        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new IntegerPushExpression(this, cloner);
        }
    }

    public class FloatPushExpression : PushExpression<double>
    {
        public FloatPushExpression(double value) : base(value)
        { }

        // copy constructor
        public FloatPushExpression(FloatPushExpression expression, Cloner cloner) : base(expression, cloner)
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack);
        }

        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new FloatPushExpression(this, cloner);
        }
    }

    public class BooleanPushExpression : PushExpression<bool>
    {
        public BooleanPushExpression(bool value) : base(value)
        { }

        // copy constructor
        public BooleanPushExpression(BooleanPushExpression expression, Cloner cloner) : base(expression, cloner)
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.BooleanStack);
        }

        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new BooleanPushExpression(this, cloner);
        }
    }

    public class NamePushExpression : PushExpression<string>
    {
        public NamePushExpression(string value) : base(value)
        { }

        // copy constructor
        public NamePushExpression(NamePushExpression expression, Cloner cloner) : base(expression, cloner)
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.NameStack);
        }

        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new NamePushExpression(this, cloner);
        }
    }

    public class ExecPushExpression : PushExpression<Expression>
    {
        public ExecPushExpression(Expression value) : base(value)
        { }

        // copy constructor
        public ExecPushExpression(ExecPushExpression expression, Cloner cloner) : base(expression, cloner)
        { }

        public override void Eval(IInterpreter interpreter)
        {
            this.Eval(interpreter.ExecStack);
        }

        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new ExecPushExpression(this, cloner);
        }
    }
}
